# from pathlib import Path
# from typing import Optional
#
# from pydantic import AnyUrl
#
# from phidata.asset.file import File
# from phidata.utils.cli_console import print_info, print_error, print_warning
# from phidata.utils.log import logger
# from phidata.task.python_task import PythonTask, PythonTaskArgs
#
#
# class DownloadUrlToFileArgs(PythonTaskArgs):
#     file: File
#     url: str
#
#
# class DownloadUrlToFile(PythonTask):
#     def __init__(
#         self,
#         file: File,
#         url: str,
#         name: str = "download_url_to_file",
#         task_id: Optional[str] = None,
#         dag_id: Optional[str] = None,
#         version: Optional[str] = None,
#         enabled: bool = True,
#     ):
#         super().__init__()
#         try:
#             self.args: DownloadUrlToFileArgs = DownloadUrlToFileArgs(
#                 file=file,
#                 url=url,
#                 name=name,
#                 task_id=task_id,
#                 dag_id=dag_id,
#                 version=version,
#                 enabled=enabled,
#                 entrypoint=download_url_to_file,
#             )
#         except Exception as e:
#             logger.error(f"Args for {self.__class__.__name__} are not valid")
#             raise
#
#     @property
#     def file(self) -> Optional[File]:
#         return self.args.file
#
#     @file.setter
#     def file(self, file: File) -> None:
#         if file is not None:
#             self.args.file = file
#
#     @property
#     def url(self) -> str:
#         return self.args.url
#
#     @url.setter
#     def url(self, url: str) -> None:
#         if url is not None:
#             self.args.url = url
#
#
# def download_url_to_file(**kwargs) -> bool:
#
#     import httpx
#     import rich.progress
#
#     args = DownloadUrlToFileArgs.from_kwargs(kwargs)
#     # logger.info(f"DownloadUrlToFileArgs: {args}")
#
#     file_to_load: Optional[File] = args.file
#     if file_to_load is None:
#         print_error("File not available")
#         return False
#     file_path: Optional[Path] = file_to_load.file_path
#     if file_path is None:
#         print_error("FilePath not available")
#         return False
#     file_dir: Path = file_path.parent.resolve()
#
#     url = args.url
#     if url is None:
#         print_error("URL not available")
#         return False
#
#     # Download URL contents
#     print_info("Downloading")
#     print_info(f"  Url: {url}")
#     print_info(f"  To : {file_path}")
#
#     # Create the directory if it does not exist
#     if not file_dir.exists():
#         file_dir.mkdir(parents=True, exist_ok=True)
#
#     bytes_to_download: int = 0
#     try:
#         with file_path.open("w") as open_file:
#             with httpx.stream("GET", url) as response:
#                 try:
#                     headers: httpx.Headers = response.headers
#                     # logger.debug("headers: {}".format(headers))
#                     # logger.debug("status_code: {}".format(response.status_code))
#                     bytes_to_download = int(headers["Content-Length"])
#                     # logger.debug("bytes_to_download : {}".format(bytes_to_download))
#                 except Exception:
#                     print_warning(f"Could not get total bytes_to_download from headers")
#
#                 with rich.progress.Progress(
#                     "[progress.percentage]{task.percentage:>3.0f}%",
#                     rich.progress.BarColumn(50),
#                     rich.progress.DownloadColumn(),
#                     rich.progress.TransferSpeedColumn(),
#                 ) as progress:
#                     download_task = progress.add_task(
#                         "Download", total=bytes_to_download
#                     )
#                     for chunk in response.iter_text():
#                         open_file.write(chunk)
#                         progress.update(
#                             download_task, completed=response.num_bytes_downloaded
#                         )
#         logger.info("Url downloaded")
#         return True
#     except Exception as e:
#         print_error("Could not download url, please try again")
#         print_info("--- stacktrace ---")
#         logger.exception(e)
#         print_info("--- stacktrace ---")
#     return False
#
#
# def best_guess_file_name(url: AnyUrl, file_name: Optional[str] = None) -> Optional[str]:
#
#     from phidata.utils.dttm import get_today_utc_date_str
#
#     fn: Optional[str] = file_name
#     if fn is None:
#         _url_path: str = str(url)
#         if "/" in _url_path:
#             fn = _url_path.split("/")[-1]
#     if fn is None:
#         fn = get_today_utc_date_str()
#     # logger.debug("fn: {}".format(fn))
#     return fn



from typing import Optional, Any, Literal

from phidata.checks.check import Check
from phidata.utils.log import logger


class NotEmpty(Check):
    def __init__(
        self,
        name: str = "TableNotEmpty",
        on_fail: Literal["fail", "warn", "ignore"] = "fail",
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__(
            name=name,
            on_fail=on_fail,
            version=version,
            enabled=enabled,
        )

    def _check_table(self, table: Any, **kwargs) -> bool:
        logger.info(f"-*- Running Check: {self.name}")
        if table is None:
            logger.error(f"Table is empty")
            return False
        return True



from typing import Optional, Any, List
from typing_extensions import Literal

from phidata.table.s3.s3_table import S3Table, S3TableFormat
from phidata.checks.check import Check
from phidata.aws.resource.s3.bucket import S3Bucket


class ParquetTableS3(S3Table):
    def __init__(
        self,
        # Table Name: required
        name: str,
        # Database for the table
        database: str = "default",
        # DataModel for this table
        data_model: Optional[Any] = None,
        # Checks to run before reading from disk
        read_checks: Optional[List[Check]] = None,
        # Checks to run before writing to disk
        write_checks: Optional[List[Check]] = None,
        # -*- Table Path
        # S3 Bucket
        bucket: Optional[S3Bucket] = None,
        # S3 Bucket Name
        bucket_name: Optional[str] = None,
        # S3 Full Path URI
        path: Optional[str] = None,
        # -*- Build S3 Path using
        top_level_dir: Optional[str] = "tables",
        path_prefix: Optional[str] = None,
        # A template string used to generate basenames of written data files.
        # The token ‘{i}’ will be replaced with an automatically incremented integer.
        # If not specified, it defaults to “part-{i}.” + format.default_extname
        basename_template: Optional[str] = None,
        # List of partition columns
        partitions: Optional[List[str]] = None,
        # Maximum number of partitions any batch may be written into.
        max_partitions: Optional[int] = None,
        # If greater than 0 then this will limit the maximum number of files that can be left open.
        # If an attempt is made to open too many files then the least recently used file will be closed.
        # If this setting is set too low you may end up fragmenting your data into many small files.
        max_open_files: Optional[int] = None,
        # Maximum number of rows per file. If greater than 0 then this will limit how many rows are placed in any single
        # file. Otherwise there will be no limit and one file will be created in each output directory unless files need
        # to be closed to respect max_open_files
        max_rows_per_file: Optional[int] = None,
        # Minimum number of rows per group. When the value is greater than 0, the dataset writer will batch incoming data
        # and only write the row groups to the disk when sufficient rows have accumulated.
        min_rows_per_group: Optional[int] = None,
        # Maximum number of rows per group. If the value is greater than 0, then the dataset writer may split up large
        # incoming batches into multiple row groups. If this value is set, then min_rows_per_group should also be set.
        # Otherwise it could end up with very small row groups.
        max_rows_per_group: Optional[int] = None,
        # Controls how the dataset will handle data that already exists in the destination.
        # The default behavior (‘error’) is to raise an error if any data exists in the destination.
        # ‘overwrite_or_ignore’ will ignore any existing data and will overwrite files with the same name
        # as an output file. Other existing files will be ignored. This behavior, in combination with a unique
        # basename_template for each write, will allow for an append workflow.
        # ‘delete_matching’ is useful when you are writing a partitioned dataset.
        # The first time each partition directory is encountered the entire directory will be deleted.
        # This allows you to overwrite old partitions completely.
        write_mode: Literal[
            "delete_matching", "overwrite_or_ignore", "error"
        ] = "delete_matching",
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            name=name,
            table_format=S3TableFormat.PARQUET,
            database=database,
            data_model=data_model,
            read_checks=read_checks,
            write_checks=write_checks,
            bucket=bucket,
            bucket_name=bucket_name,
            path=path,
            top_level_dir=top_level_dir,
            path_prefix=path_prefix,
            basename_template=basename_template,
            partitions=partitions,
            max_partitions=max_partitions,
            max_open_files=max_open_files,
            max_rows_per_file=max_rows_per_file,
            min_rows_per_group=min_rows_per_group,
            max_rows_per_group=max_rows_per_group,
            write_mode=write_mode,
            version=version,
            enabled=enabled,
            **kwargs,
        )



from typing import Optional, Any, Literal

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.utils.log import logger


class CheckArgs(PhidataBaseArgs):
    # Check Name
    name: str
    # How to handle check failure
    on_fail: Literal["fail", "warn", "ignore"] = "fail"


class Check(PhidataBase):
    """Base Class for all Checks"""

    def __init__(
        self,
        name: str,
        on_fail: Literal["fail", "warn", "ignore"] = "fail",
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: CheckArgs = CheckArgs(
                name=name,
                on_fail=on_fail,
                version=version,
                enabled=enabled,
            )
        except Exception:
            logger.error(f"Args for {self.name} are not valid")
            raise

    @property
    def on_fail(self) -> Literal["fail", "warn", "ignore"]:
        return self.args.on_fail if self.args else "fail"

    @on_fail.setter
    def on_fail(self, on_fail: Literal["fail", "warn", "ignore"]) -> None:
        if self.args and on_fail in ["fail", "warn", "ignore"]:
            self.args.on_fail = on_fail

    def validate(self, result: bool = False) -> bool:
        if result:
            logger.info(f"-*- Check: {self.name} passed")
        else:
            if self.args.on_fail == "fail":
                raise Exception(f"Check: {self.name} failed")
            elif self.args.on_fail == "warn":
                logger.warning(f"Check: {self.name} failed")
            else:
                logger.info(f"-*- Check: {self.name} failed")

        # Blocking/failed checks return an Exception above
        return True

    def _check_polars_df(self, df: Any, **kwargs) -> bool:
        # subclasses should implement this check and return True/False
        logger.error(f"@check_dataframe not defined for {self.name}")
        return True

    def check_polars_df(self, df: Any, **kwargs) -> bool:
        return self.validate(self._check_polars_df(df, **kwargs))

    def _check_pandas_df(self, df: Any, **kwargs) -> bool:
        # subclasses should implement this check and return True/False
        logger.error(f"@check_dataframe not defined for {self.name}")
        return True

    def check_pandas_df(self, df: Any, **kwargs) -> bool:
        return self.validate(self._check_pandas_df(df, **kwargs))

    def _check_table(self, table: Any, **kwargs) -> bool:
        # subclasses should implement this check and return True/False
        logger.error(f"@check_table not defined for {self.name}")
        return True

    def check_table(self, table: Any, **kwargs) -> bool:
        return self.validate(self._check_table(table, **kwargs))



from pathlib import Path
from typing import Optional, Union, Any, Dict
from typing_extensions import Literal

from phidata.asset.local import LocalAsset, LocalAssetArgs
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class LocalFileFormat(ExtendedEnum):
    CSV = "csv"
    TSV = "tsv"
    TXT = "txt"
    JSON = "json"
    PARQUET = "parquet"
    FEATHER = "feather"


class LocalFileArgs(LocalAssetArgs):
    # File name
    name: Optional[str] = None
    # File format
    file_format: Optional[LocalFileFormat] = None

    # True If the file is located in the current dir.
    # Default: False
    current_dir: bool = False

    # Parent directory of the file relative to the storage_dir
    # Used to build the file_path
    file_dir: Optional[Union[str, Path]] = None

    # Absolute path for the file
    file_path: Optional[Path] = None


class LocalFile(LocalAsset):
    def __init__(
        self,
        # File name
        name: Optional[str] = None,
        # File format
        file_format: Optional[LocalFileFormat] = None,
        # Set as True if the file is located in the current directory.
        # By default, we store files in the storage_dir.
        # Default: False
        current_dir: bool = False,
        # Parent directory of the file relative to the storage_dir
        # Used to build the file_path
        file_dir: Optional[Union[str, Path]] = None,
        # Absolute path for the file
        file_path: Optional[Path] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ) -> None:

        super().__init__()
        try:
            self.args: LocalFileArgs = LocalFileArgs(
                name=name,
                file_format=file_format,
                current_dir=current_dir,
                file_dir=file_dir,
                file_path=file_path,
                version=version,
                enabled=enabled,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise

    @property
    def file_format(self) -> Optional[LocalFileFormat]:
        if self.args.file_format is not None:
            return self.args.file_format

        # logger.debug("-*--*- Building file_format")
        fp = self.file_path
        if fp is None:
            return None
        suffix = fp.suffix[1:]  # remove the . from ".json" or ".csv"
        # logger.debug("suffix: {}".format(suffix))

        derived_file_format: Optional[LocalFileFormat] = None
        if suffix.lower() in LocalFileFormat.values_list():
            derived_file_format = LocalFileFormat.from_str(suffix.lower())
            self.args.file_format = derived_file_format
        return self.args.file_format

    @file_format.setter
    def file_format(self, file_format: Optional[LocalFileFormat]) -> None:
        if file_format is not None:
            self.args.file_format = file_format

    @property
    def current_dir(self) -> bool:
        return self.args.current_dir if self.args else True

    @current_dir.setter
    def current_dir(self, current_dir: bool) -> None:
        if current_dir is not None:
            self.args.current_dir = current_dir

    @property
    def file_dir(self) -> Optional[Union[str, Path]]:
        return self.args.file_dir if self.args else None

    @file_dir.setter
    def file_dir(self, file_dir: Union[str, Path]) -> None:
        if file_dir is not None:
            self.args.file_dir = file_dir

    @property
    def file_path(self) -> Optional[Path]:
        if self.args.file_path is not None:
            return self.args.file_path

        logger.debug("-*--*- Building file_path")
        _file_path: Optional[Path] = None

        # Use current_dir as base path if set
        if self.current_dir:
            logger.debug("use current_dir: {}".format(self.current_dir))
            _file_path = Path(__file__).resolve()

        # Or use storage_dir_path as the base path
        if _file_path is None:
            # storage_dir_path is loaded from the current environment variable
            _file_path = self.storage_dir_path

        # Add the file_dir if available
        if self.file_dir is not None:
            if _file_path is None:
                # _file_path is None meaning no storage_dir_path
                _file_path = Path(".").resolve().joinpath(self.file_dir)
            else:
                _file_path = _file_path.joinpath(self.file_dir)

        # Add the file_name
        if self.name is not None:
            if _file_path is None:
                _file_path = Path(".").resolve().joinpath(self.name)
            else:
                _file_path = _file_path.joinpath(self.name)

        self.args.file_path = _file_path
        return self.args.file_path

    @classmethod
    def temporary(cls, name: str) -> "LocalFile":
        return cls(name=name, file_dir="tmp")

    ######################################################
    ## Write to file
    ######################################################

    def write_polars_df(
        self,
        # A polars DataFrame
        df: Optional[Any] = None,
        #
        # -*- CSV Files
        #
        # Whether to include header in the CSV output.
        include_header: Optional[bool] = None,
        # Separate CSV fields with this symbol.
        sep: Optional[str] = None,
        # Byte to use as quoting character.
        quote: Optional[str] = None,
        # Number of rows that will be processed per thread.
        batch_size: Optional[int] = None,
        datetime_format: Optional[str] = None,
        date_format: Optional[str] = None,
        time_format: Optional[str] = None,
        float_precision: Optional[int] = None,
        null_value: Optional[str] = None,
        #
        # -*- JSON Files
        #
        # Pretty serialize json.
        pretty: Optional[bool] = None,
        # Write to row oriented json. This is slower, but more common.
        row_oriented: Optional[bool] = None,
        # -*- How to behave if the file already exists.
        # fail: Raise a ValueError.
        # replace: Drop the file before inserting new values.
        # append: Insert new values to the existing table.
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        **kwargs,
    ) -> bool:
        """
        Write Polars DataFrame to a file.
        """

        # LocalFile not yet initialized
        if self.args is None:
            return False

        # Check path is available
        if self.file_path is None:
            logger.error("FilePath invalid")
            return False

        # Validate pyarrow is installed
        try:
            import pyarrow as pa
        except ImportError:
            logger.error("PyArrow not installed")
            return False

        # Validate polars is installed
        try:
            import polars as pl  # type: ignore
        except ImportError:
            logger.error("Polars not installed")
            return False

        if df is None or not isinstance(df, pl.DataFrame):
            logger.error("DataFrame invalid")
            return False

        # Write the DataFrame to a file
        file_path: Path = self.file_path
        logger.info("Writing to file: {}".format(file_path))
        if file_path.exists():
            if if_exists == "fail":
                raise ValueError("File exists: {}".format(file_path))
            elif if_exists == "replace":
                from phidata.utils.filesystem import delete_from_fs

                logger.debug("Deleting: {}".format(file_path))
                delete_from_fs(file_path)

        # Create parent directories if they don't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            logger.debug("Format: {}".format(self.file_format))
            if self.file_format == LocalFileFormat.CSV:
                # Create a dict of args which are not null
                csv_args: Dict[str, Any] = {}
                if include_header is not None:
                    csv_args["include_header"] = include_header
                if sep is not None:
                    csv_args["sep"] = sep
                if quote is not None:
                    csv_args["quote"] = quote
                if batch_size is not None:
                    csv_args["batch_size"] = batch_size
                if datetime_format is not None:
                    csv_args["datetime_format"] = datetime_format
                if date_format is not None:
                    csv_args["date_format"] = date_format
                if time_format is not None:
                    csv_args["time_format"] = time_format
                if float_precision is not None:
                    csv_args["float_precision"] = float_precision
                if null_value is not None:
                    csv_args["null_value"] = null_value

                df.write_csv(file_path, **csv_args)  # type: ignore
            elif self.file_format == LocalFileFormat.JSON:
                # Create a dict of args which are not null
                json_args: Dict[str, Any] = {}
                if include_header is not None:
                    json_args["pretty"] = pretty
                if sep is not None:
                    json_args["row_oriented"] = row_oriented

                df.write_json(file_path, **json_args)  # type: ignore
            else:
                logger.error(f"FileFormat: {self.file_format} not yet supported")
                return False
            return True
        except Exception as e:
            logger.error("Could not write to: {}".format(file_path))
            raise

    def write_pandas_df(
        self,
        df: Any = None,
        # How to behave if the file already exists.
        # fail: Raise a ValueError.
        # replace: Drop the file before inserting new values.
        # append: Insert new values to the existing table.
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        **kwargs,
    ) -> bool:
        """
        Write DataFrame to file.
        """

        # LocalFile not yet initialized
        if self.args is None:
            return False

        # Check path is available
        if self.file_path is None:
            logger.error("FilePath invalid")
            return False

        # Validate pandas is installed
        try:
            import pandas as pd
        except ImportError:
            logger.error("Pandas not installed")
            return False

        if df is None or not isinstance(df, pd.DataFrame):
            logger.error("DataFrame invalid")
            return False

        file_path: Path = self.file_path
        logger.info("Writing to file: {}".format(file_path))
        mode: str = "w"
        if file_path.exists():
            if if_exists == "fail":
                raise ValueError("LocalFile already exists: {}".format(file_path))
            elif if_exists == "replace":
                from phidata.utils.filesystem import delete_from_fs

                logger.debug("Deleting: {}".format(file_path))
                delete_from_fs(file_path)
            elif if_exists == "append":
                mode = "a"

        # Create parent directories if they don't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            file_path.touch(exist_ok=True)
            with file_path.open(mode) as open_file:
                logger.debug("file_format: {}".format(self.file_format))
                if self.file_format == LocalFileFormat.JSON:
                    df.to_json(open_file, orient="index", indent=4)
                elif self.file_format == LocalFileFormat.CSV:
                    df.to_csv(open_file)
                else:
                    logger.error(
                        f"LocalFileFormat: {self.file_format} not yet supported"
                    )
                    return False
            return True
        except Exception as e:
            logger.error("Could not write to: {}".format(file_path))
            raise

    def download_url(self, url: str) -> bool:
        # LocalFile not yet initialized
        if self.args is None:
            return False

        import httpx
        import rich.progress

        file_path: Optional[Path] = self.file_path
        if file_path is None:
            logger.error("FilePath not available")
            return False
        file_dir: Path = file_path.parent.resolve()

        # Download URL contents
        logger.info("Downloading")
        logger.info(f"  Url: {url}")
        logger.info(f"  To : {file_path}")

        # Create the directory if it does not exist
        if not file_dir.exists():
            file_dir.mkdir(parents=True, exist_ok=True)

        bytes_to_download: int = 0
        try:
            with file_path.open("w") as open_file:
                with httpx.stream("GET", url) as response:
                    try:
                        headers: httpx.Headers = response.headers
                        # logger.debug("headers: {}".format(headers))
                        # logger.debug("status_code: {}".format(response.status_code))
                        bytes_to_download = int(headers["Content-Length"])
                        # logger.debug("bytes_to_download : {}".format(bytes_to_download))
                    except Exception:
                        logger.warning(
                            f"Could not get total bytes_to_download from headers"
                        )

                    with rich.progress.Progress(
                        "[progress.percentage]{task.percentage:>3.0f}%",
                        rich.progress.BarColumn(50),
                        rich.progress.DownloadColumn(),
                        rich.progress.TransferSpeedColumn(),
                    ) as progress:
                        download_task = progress.add_task(
                            "Download", total=bytes_to_download
                        )
                        for chunk in response.iter_text():
                            open_file.write(chunk)
                            progress.update(
                                download_task, completed=response.num_bytes_downloaded
                            )
            logger.info("Url downloaded")
            return True
        except Exception as e:
            logger.error("Could not download url, please try again")
            logger.exception(e)
        return False



from typing import Any, Dict, List, Optional

from rich.console import Console

console = Console()


def print_table(
    title: Optional[str] = None,
    header: Optional[List[str]] = None,
    rows: Optional[List[Dict[str, Any]]] = None,
) -> None:
    from rich import box
    from rich.table import Table

    table = Table(
        title=title,
        box=box.ASCII2,
        show_lines=True,
    )

    if header:
        for col in header:
            table.add_column(col, justify="center", no_wrap=True)

    if rows:
        for row in rows:
            row_values = [str(v) for v in row.values()]
            table.add_row(*row_values)

    if table.row_count > 0:
        console.print(table)



from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Optional, Any, List, Dict, Literal

from phidata.asset.data_asset import DataAsset
from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.checks.check import Check
from phidata.exceptions.workflow import WorkflowFailure
from phidata.utils.context import get_run_date, build_path_context_from_env
from phidata.utils.cli_console import print_info, print_subheading
from phidata.utils.log import logger
from phidata.types.run_status import RunStatus
from phidata.types.context import PathContext, RunContext, AirflowContext
from phidata.workflow import Workflow
from phidata.workflow.workflow_relatives import WorkflowRelatives
from phidata.utils.env_var import validate_env_vars


class DataProductArgs(PhidataBaseArgs):
    # List of workflows in this DataProduct
    workflows: Optional[List[Workflow]] = None
    # A dependency graph of the workflows in the format:
    # { downstream: [upstream_list] }
    # The downstream workflow will execute after all workflows in the
    #   upstream_list are finished.
    # upstream_list: workflows which run before the downstream workflow
    # Ex: To run extract -> transform -> load
    #   graph = {
    #     transform: [extract],
    #     load: [transform],
    #     ...
    #   }
    graph: Optional[Dict[Workflow, List[Workflow]]] = None
    # What to do when a workflow fails while running the data_product locally.
    # Continue to next task or stop?
    on_fail: Literal["stop", "continue"] = "stop"

    # Add an EmptyOperator at the start of the workflows
    add_start_task: bool = True
    # Add an EmptyOperator at the end of the workflows
    add_end_task: bool = False

    # airflow dag_id for this data_product, use name if not provided
    dag_id: Optional[str] = None
    # run_context provided by wf_operator
    run_context: Optional[RunContext] = None
    # path_context provided by env variables
    path_context: Optional[PathContext] = None
    # airflow_context provided by airflow
    airflow_context: Optional[AirflowContext] = None
    # Env vars that validate Airflow is active on the containers
    # This is used to gate code blocks which should only run on
    # remote containers like creating DAGs and tasks
    validate_airflow_env: Dict[str, Any] = {"INIT_AIRFLOW": True}

    # Input DataAssets used to build sensors before kicking off the data_product
    inputs: Optional[List[DataAsset]] = None
    # DataAssets produced by this data_product, used for building the lineage graph
    outputs: Optional[List[DataAsset]] = None

    # Checks to run before the data_product
    pre_checks: Optional[List[Check]] = None
    # Checks to run after the data_product
    post_checks: Optional[List[Check]] = None

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    @property
    def run_date(self) -> datetime:
        return get_run_date(self.run_context, self.airflow_context)

    @classmethod
    def from_kwargs(cls, kwargs: Optional[Dict] = None):
        if kwargs is None or not isinstance(kwargs, dict):
            return cls()
        # logger.info(f"Loading {cls.__name__} using kwargs")
        args_object = cls(**kwargs)
        validate_airflow_env = kwargs.get("validate_airflow_env", None)
        if validate_env_vars(validate_airflow_env):
            # logger.info("Creating airflow_context")
            airflow_context = AirflowContext(**kwargs)
            args_object.airflow_context = airflow_context
        return args_object


class DataProduct(PhidataBase):
    """Base Class for all DataProducts"""

    def __init__(
        self,
        name: str,
        # List of workflows in this DataProduct,
        workflows: Optional[List[Workflow]] = None,
        # A dependency graph of the workflows in the format:
        # { downstream: [upstream_list] }
        # The downstream workflow will execute after all workflows in the
        #   upstream_list are finished.
        # upstream_list: workflows which run before the downstream workflow
        # Ex: To run extract -> transform -> load
        #   graph = {
        #     transform: [extract],
        #     load: [transform],
        #     ...
        #   }
        graph: Optional[Dict[Workflow, List[Workflow]]] = None,
        # What to do when a task fails while running the workflow locally.
        # Continue to next task or stop?
        on_fail: Literal["stop", "continue"] = "stop",
        # Add an EmptyOperator at the start of the workflows
        add_start_task: bool = True,
        # Add an EmptyOperator at the end of the workflows
        add_end_task: bool = False,
        # airflow dag_id for this workflow, use name if not provided
        dag_id: Optional[str] = None,
        # run_context provided by wf_operator
        run_context: Optional[RunContext] = None,
        # path_context provided by env variables
        path_context: Optional[PathContext] = None,
        # airflow_context provided by airflow
        airflow_context: Optional[AirflowContext] = None,
        # Input DataAssets used to build sensors before kicking off the workflow
        inputs: Optional[List[DataAsset]] = None,
        # DataAssets produced by this workflow, used for building the lineage graph
        outputs: Optional[List[DataAsset]] = None,
        # Checks to run before the workflow
        pre_checks: Optional[List[Check]] = None,
        # Checks to run after the workflow
        post_checks: Optional[List[Check]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:
        super().__init__()
        self._family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = None
        try:
            self.args: DataProductArgs = DataProductArgs(
                name=name,
                workflows=workflows,
                graph=graph,
                on_fail=on_fail,
                add_start_task=add_start_task,
                add_end_task=add_end_task,
                dag_id=dag_id,
                run_context=run_context,
                path_context=path_context,
                airflow_context=airflow_context,
                inputs=inputs,
                outputs=outputs,
                pre_checks=pre_checks,
                post_checks=post_checks,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def workflows(self) -> Optional[List[Workflow]]:
        return self.args.workflows if self.args else None

    @workflows.setter
    def workflows(self, workflows: List[Workflow]) -> None:
        if self.args is not None and workflows is not None:
            self.args.workflows = workflows

    @property
    def graph(self) -> Optional[Dict[Workflow, List[Workflow]]]:
        return self.args.graph if self.args else None

    @graph.setter
    def graph(self, graph: Dict[Workflow, List[Workflow]]) -> None:
        if self.args is not None and graph is not None:
            self.args.graph = graph

    @property
    def on_fail(self) -> Literal["stop", "continue"]:
        return self.args.on_fail if self.args else "stop"

    @on_fail.setter
    def on_fail(self, on_fail: Literal["stop", "continue"]) -> None:
        if self.args is not None and on_fail is not None:
            self.args.on_fail = on_fail

    @property
    def add_start_task(self) -> bool:
        return self.args.add_start_task if self.args else False

    @add_start_task.setter
    def add_start_task(self, add_start_task: bool) -> None:
        if self.args is not None and add_start_task is not None:
            self.args.add_start_task = add_start_task

    @property
    def add_end_task(self) -> bool:
        return self.args.add_end_task if self.args else False

    @add_end_task.setter
    def add_end_task(self, add_end_task: bool) -> None:
        if self.args is not None and add_end_task is not None:
            self.args.add_end_task = add_end_task

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        # Workflow not yet initialized
        if self.args is None:
            return None

        if self.args.path_context is not None:
            # use cached value if available
            return self.args.path_context

        self.args.path_context = build_path_context_from_env()
        return self.args.path_context

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def airflow_context(self) -> Optional[AirflowContext]:
        return self.args.airflow_context if self.args else None

    @airflow_context.setter
    def airflow_context(self, airflow_context: AirflowContext) -> None:
        if self.args is not None and airflow_context is not None:
            self.args.airflow_context = airflow_context

    @property
    def inputs(self) -> Optional[List[DataAsset]]:
        return self.args.inputs if self.args else None

    @inputs.setter
    def inputs(self, inputs: List[DataAsset]) -> None:
        if self.args is not None and inputs is not None:
            self.args.inputs = inputs

    @property
    def outputs(self) -> Optional[List[DataAsset]]:
        return self.args.outputs if self.args else None

    @outputs.setter
    def outputs(self, outputs: List[DataAsset]) -> None:
        if self.args is not None and outputs is not None:
            self.args.outputs = outputs

    @property
    def pre_checks(self) -> Optional[List[Check]]:
        return self.args.pre_checks if self.args else None

    @pre_checks.setter
    def pre_checks(self, pre_checks: List[Check]) -> None:
        if self.args is not None and pre_checks is not None:
            self.args.pre_checks = pre_checks

    @property
    def post_checks(self) -> Optional[List[Check]]:
        return self.args.post_checks if self.args else None

    @post_checks.setter
    def post_checks(self, post_checks: List[Check]) -> None:
        if self.args is not None and post_checks is not None:
            self.args.post_checks = post_checks

    ######################################################
    ## DataProduct Properties
    ######################################################

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.args.validate_airflow_env)

    @property
    def family_tree(self) -> Optional[Dict[Workflow, WorkflowRelatives]]:
        if self.workflows is None:
            return None

        # used cached value if available
        if self._family_tree is not None:
            return self._family_tree

        # build family_tree
        family_tree: Dict[Workflow, WorkflowRelatives] = OrderedDict()
        for workflow in self.workflows:
            family_tree[workflow] = WorkflowRelatives()

        if self.graph is None:
            return family_tree

        for workflow, upstream_list in self.graph.items():
            family_tree[workflow].add_upstream_list(upstream_list)
            for upstream_workflow in upstream_list:
                family_tree[upstream_workflow].add_downstream(workflow)

        return family_tree

    @property
    def family_tree_dict(self) -> Dict[str, Dict[str, List[str]]]:
        tree: Dict[str, Dict[str, List[str]]] = OrderedDict()
        family_tree = self.family_tree
        if family_tree is None:
            return tree

        for workflow in family_tree:
            workflow_tree = {}
            upstream_list = family_tree[workflow].upstream_list
            downstream_list = family_tree[workflow].downstream_list
            if upstream_list:
                workflow_tree["upstream"] = [wf.name for wf in upstream_list]
            if downstream_list:
                workflow_tree["downstream"] = [wf.name for wf in downstream_list]
            tree[workflow.name] = workflow_tree
        return tree

    @property
    def roots(self) -> Optional[List[Workflow]]:
        """
        Roots are workflows that run first. i.e. have nothing upstream of them
        """
        if self.graph is None:
            return self.workflows

        family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = self.family_tree
        if family_tree is None:
            return self.workflows

        roots: List[Workflow] = []
        for workflow, relatives in family_tree.items():
            if relatives.upstream_list is None:
                roots.append(workflow)
        return roots

    @property
    def leaves(self) -> Optional[List[Workflow]]:
        """
        Leaves are workflows that run last. i.e. have nothing downstream of them
        """
        if self.graph is None:
            return self.workflows

        family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = self.family_tree
        if family_tree is None:
            return self.workflows

        leaves: List[Workflow] = []
        for workflow, relatives in family_tree.items():
            if relatives.downstream_list is None:
                leaves.append(workflow)
        return leaves

    def __call__(self, *args, **kwargs) -> bool:
        logger.warning("DataProducts should not be called directly")
        return True

    ######################################################
    ## Build data_product
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Workflows
    ######################################################

    def add_workflow(self, workflow: Workflow):
        if self.args.workflows is None:
            self.args.workflows = []
        self.args.workflows.append(workflow)

    ######################################################
    ## Checks
    ######################################################

    def add_pre_check(self, check: Check):
        if self.args.pre_checks is None:
            self.args.pre_checks = []
        self.args.pre_checks.append(check)

    def add_post_check(self, check: Check):
        if self.args.post_checks is None:
            self.args.post_checks = []
        self.args.post_checks.append(check)

    ######################################################
    ## Run data_product
    ######################################################

    def run_workflow_in_local_env(
        self, workflow: Workflow, task_id: Optional[str] = None
    ) -> RunStatus:
        wf_name = workflow.name
        print_subheading(f"\nRunning workflow: {wf_name}")
        # Pass down context
        workflow.run_context = self.run_context
        workflow.path_context = self.path_context
        workflow.airflow_context = self.airflow_context
        # Not required for local runs but adding for posterity
        if self.dag_id:
            workflow.dag_id = self.dag_id
        run_success = workflow.run_in_local_env(task_id=task_id)
        return RunStatus(name=wf_name, success=run_success)

    def run_in_local_env(
        self, workflow_name: Optional[str] = None, task_id: Optional[str] = None
    ) -> bool:
        """
        Runs a DataProduct in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running DataProduct locally")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        if workflow_name is not None:
            wf_name_map = {wf.name: wf for wf in self.workflows}
            if workflow_name in wf_name_map:
                wf_run_status.append(
                    self.run_workflow_in_local_env(
                        workflow=wf_name_map[workflow_name], task_id=task_id
                    )
                )
            else:
                logger.error(f"Could not find {workflow_name} in {wf_name_map.keys()}")
        else:
            for wf in self.workflows:
                run_status = self.run_workflow_in_local_env(wf)
                wf_run_status.append(run_status)
                if not run_status.success and self.args.on_fail == "stop":
                    # logger.error(f"Workflow {wf.name} failed")
                    break

        print_run_status_table("Workflow Run Status", wf_run_status)
        for run_status in wf_run_status:
            if not run_status.success:
                return False
        return True

    def run_workflow_in_docker_container(
        self,
        workflow: Workflow,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
        task_id: Optional[str] = None,
    ) -> RunStatus:
        _wf_name = workflow.name
        print_subheading(f"\nRunning workflow: {_wf_name}")
        # Pass down context
        workflow.run_context = self.run_context
        workflow.path_context = self.path_context
        workflow.airflow_context = self.airflow_context
        workflow.dag_id = self.dag_id
        run_success = workflow.run_in_docker_container(
            active_container=active_container, docker_env=docker_env, task_id=task_id
        )
        return RunStatus(name=_wf_name, success=run_success)

    def run_in_docker_container(
        self,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
        workflow_name: Optional[str] = None,
        task_id: Optional[str] = None,
    ) -> bool:
        """
        Runs a DataProduct in a docker container.

        Args:
            active_container:
            docker_env:
            workflow_name:
            task_id:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running DataProduct in docker container")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        if workflow_name is not None:
            wf_name_map = {wf.name: wf for wf in self.workflows}
            if workflow_name in wf_name_map:
                wf_run_status.append(
                    self.run_workflow_in_docker_container(
                        workflow=wf_name_map[workflow_name],
                        active_container=active_container,
                        docker_env=docker_env,
                        task_id=task_id,
                    )
                )
            else:
                logger.error(f"Could not find {workflow_name} in {wf_name_map.keys()}")
        else:
            for wf in self.workflows:
                run_status = self.run_workflow_in_docker_container(
                    workflow=wf,
                    active_container=active_container,
                    docker_env=docker_env,
                )
                wf_run_status.append(run_status)
                if not run_status.success and self.args.on_fail == "stop":
                    raise WorkflowFailure(f"Workflow {wf.name} failed")

        print_run_status_table("Workflow Run Status", wf_run_status)
        for run_status in wf_run_status:
            if not run_status.success:
                return False
        return True

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a DataProduct in a k8s container.

        Args:
            pod: The pod to run the workflow in
            k8s_api_client: K8sApiClient
            container_name: The container to run the workflow in
            k8s_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the k8s container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        logger.info("--**-- Running DataProduct in k8s container")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        for idx, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, idx)
            print_subheading(f"\nRunning {wf_name}")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            wf.airflow_context = self.airflow_context
            wf.dag_id = self.dag_id
            run_success = wf.run_in_k8s_container(
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
            wf_run_status.append(RunStatus(name=wf_name, success=run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    ######################################################
    ## Airflow functions
    ######################################################

    def create_airflow_dag(
        self,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        # Default: daily
        schedule_interval: timedelta = timedelta(days=1),
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # if start_date is not provided, use start_days_ago
        start_days_ago: int = 1,
        # A date beyond which your DAG won't run, leave to None
        #  for open ended scheduling
        end_date: Optional[datetime] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        concurrency: Optional[int] = None,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: int = 8,
        doc_md: Optional[str] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        is_paused_upon_creation: Optional[bool] = None,
        jinja_environment_kwargs: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Used to create an Airflow DAG for the DataProduct.
        """

        # Skips DAG creation on local machines
        if not self.airflow_active:
            return None

        from airflow import DAG
        from airflow.models.operator import Operator
        from airflow.utils.dates import days_ago

        operator_default_args = default_args
        if operator_default_args is None:
            operator_default_args = {}
        operator_default_args.update(
            {
                "owner": owner,
                "depends_on_past": depends_on_past,
            }
        )
        if start_date is None:
            start_date = days_ago(start_days_ago)

        dag_id = self.dag_id
        if dag_id is None:
            logger.error("DataProduct dag_id unavailable")
            return None
        if self.workflows is None:
            logger.error(f"DataProduct has no workflows")
            return None
        dag = DAG(
            dag_id=dag_id,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
            user_defined_macros=user_defined_macros,
            user_defined_filters=user_defined_filters,
            default_args=operator_default_args,
            concurrency=concurrency,
            max_active_runs=max_active_runs,
            doc_md=doc_md,
            params=params,
            is_paused_upon_creation=is_paused_upon_creation,
            jinja_environment_kwargs=jinja_environment_kwargs,
            tags=tags,
        )

        # Add airflow tasks from workflows to DAG
        for workflow in self.workflows:
            # Add start tasks to workflows if there are inter-workflow dependencies
            if self.add_start_task or self.graph is not None:
                workflow.add_start_task = True
            if self.add_end_task or self.graph is not None:
                workflow.add_end_task = True
            # logger.info(f"Adding tasks for workflow: {workflow.name}")
            add_task_success = workflow.add_airflow_tasks_to_dag(dag=dag)
            if not add_task_success:
                logger.error(f"Tasks for workflow {workflow.name} could not be added")

        if self.graph is not None:
            # logger.info("Updating dependencies")
            for downstream_workflow, upstream_workflow_list in self.graph.items():

                downstream_wf_start_task: Optional[Operator] = None
                downstream_wf_start_task_id = downstream_workflow.start_workflow_task_id
                # logger.info(f"downstream_workflow_start: {downstream_wf_start_task_id}")
                if dag.has_task(downstream_wf_start_task_id):
                    downstream_wf_start_task = dag.get_task(downstream_wf_start_task_id)
                if downstream_wf_start_task is None:
                    logger.info(f"Cannot find task: {downstream_wf_start_task_id}")
                    continue

                for workflow in upstream_workflow_list:
                    upstream_wf_end_task: Optional[Operator] = None
                    upstream_wf_end_task_id = workflow.end_workflow_task_id
                    # logger.info(f"upstream_workflow_end: {upstream_wf_end_task_id}")

                    if dag.has_task(upstream_wf_end_task_id):
                        upstream_wf_end_task = dag.get_task(upstream_wf_end_task_id)

                    if upstream_wf_end_task is not None:
                        downstream_wf_start_task.set_upstream(upstream_wf_end_task)

                # logger.info(f"Upstream: {downstream_wf_start_task.upstream_task_ids}")
        logger.debug(f"Airflow dag {dag_id} created")
        return dag



from pathlib import Path
from typing import Optional, List, Dict, Union, Any

from phidata.infra.args import InfraArgs
from phidata.infra.config import InfraConfig
from phidata.workspace.settings import WorkspaceSettings
from phidata.utils.log import logger


class WorkspaceConfigArgs(InfraArgs):
    # default env for phi ws ... commands
    default_env: Optional[str] = None
    # default config for phi ws ... commands
    default_config: Optional[str] = None
    # A copy of the workspace settings
    ws_settings: Optional[WorkspaceSettings] = None

    # List of Docker configurations
    # Type: DockerConfig
    docker: Optional[List[Any]] = None
    # List of K8s configurations
    # Type: K8sConfig
    k8s: Optional[List[Any]] = None
    # List of AWS configurations
    # Type: AwsConfig
    aws: Optional[List[Any]] = None

    # Path to important directories relative to the workspace_root
    meta_dir: str
    notebooks_dir: str
    products_dir: str
    scripts_dir: str
    storage_dir: str
    workflows_dir: str
    workspace_config_dir: str


class WorkspaceConfig(InfraConfig):
    def __init__(
        self,
        version: Optional[str] = None,
        enabled: bool = True,
        # default env for phi ws ... commands
        default_env: Optional[str] = None,
        # default config for phi ws ... commands
        default_config: Optional[str] = None,
        ws_settings: Optional[WorkspaceSettings] = None,
        # List of Docker configurations
        # Type: DockerConfig
        docker: Optional[List[Any]] = None,
        # List of K8s configurations
        # Type: K8sConfig
        k8s: Optional[List[Any]] = None,
        # List of AWS configurations
        # Type: AwsConfig
        aws: Optional[List[Any]] = None,
        # Path to important directories relative to the workspace_root
        meta_dir: str = "meta",
        notebooks_dir: str = "notebooks",
        products_dir: str = "workflows",
        scripts_dir: str = "scripts",
        storage_dir: str = "storage",
        workflows_dir: str = "workflows",
        workspace_config_dir: str = "workspace",
        # -*- Environment parameters
        # NOTE: These are NOT used by the WorkspaceConfig
        # So cannot be used in the workspace files
        # Env vars added to local env when running workflows
        local_env: Optional[Dict[str, str]] = None,
        # Load local env using env file
        local_env_file: Optional[Union[str, Path]] = None,
        # Env vars added to docker env when running workflows
        docker_env: Optional[Dict[str, str]] = None,
        # Load docker env using env file
        docker_env_file: Optional[Union[str, Path]] = None,
        # Env vars added to k8s env when running workflows
        k8s_env: Optional[Dict[str, str]] = None,
        # Load k8s env using env file
        k8s_env_file: Optional[Union[str, Path]] = None,
        # -*- AWS parameters
        # Common aws params used by apps, resources and data assets
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_config_file: Optional[str] = None,
        aws_shared_credentials_file: Optional[str] = None,
        # -*- `phi` cli parameters
        # Set to True if `phi` should continue creating
        # resources after a resource creation has failed
        continue_on_create_failure: bool = False,
        # Set to True if `phi` should continue deleting
        # resources after a resource deleting has failed
        # Defaults to True because we normally want to continue deleting
        continue_on_delete_failure: bool = True,
        # Set to True if `phi` should continue patching
        # resources after a resource patch has failed
        continue_on_patch_failure: bool = False,
    ):
        super().__init__()

        try:
            self.args: WorkspaceConfigArgs = WorkspaceConfigArgs(
                version=version,
                enabled=enabled,
                default_env=default_env,
                default_config=default_config,
                ws_settings=ws_settings,
                docker=docker,
                k8s=k8s,
                aws=aws,
                meta_dir=meta_dir,
                notebooks_dir=notebooks_dir,
                products_dir=products_dir,
                scripts_dir=scripts_dir,
                storage_dir=storage_dir,
                workflows_dir=workflows_dir,
                workspace_config_dir=workspace_config_dir,
                local_env=local_env,
                local_env_file=local_env_file,
                docker_env=docker_env,
                docker_env_file=docker_env_file,
                k8s_env=k8s_env,
                k8s_env_file=k8s_env_file,
                aws_region=aws_region,
                aws_profile=aws_profile,
                aws_config_file=aws_config_file,
                aws_shared_credentials_file=aws_shared_credentials_file,
                continue_on_create_failure=continue_on_create_failure,
                continue_on_delete_failure=continue_on_delete_failure,
                continue_on_patch_failure=continue_on_patch_failure,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def default_env(self) -> Optional[str]:
        return self.args.default_env if self.args else None

    @property
    def default_config(self) -> Optional[str]:
        return self.args.default_config if self.args else None

    @property
    def docker(self) -> Optional[List[Any]]:
        return self.args.docker if self.args else None

    @property
    def k8s(self) -> Optional[List[Any]]:
        return self.args.k8s if self.args else None

    @property
    def aws(self) -> Optional[List[Any]]:
        return self.args.aws if self.args else None

    def is_valid(self) -> bool:
        if self.docker is not None:
            from phidata.docker.config import DockerConfig

            if not isinstance(self.docker, list):
                raise ValueError("docker should be a list")
            for dc in self.docker:
                if not isinstance(dc, DockerConfig):
                    raise ValueError(f"Invalid DockerConfig: {dc}")

        if self.k8s is not None:
            from phidata.k8s.config import K8sConfig

            if not isinstance(self.k8s, list):
                raise ValueError("k8s should be a list")
            for kc in self.k8s:
                if not isinstance(kc, K8sConfig):
                    raise ValueError(f"Invalid K8sConfig: {kc}")

        if self.aws is not None:
            from phidata.aws.config import AwsConfig

            if not isinstance(self.aws, list):
                raise ValueError("aws should be a list")
            for awsc in self.aws:
                if not isinstance(awsc, AwsConfig):
                    raise ValueError(f"Invalid AwsConfig: {awsc}")
        return True



from typing import Optional, Any

from phidata.asset.data_asset import DataAsset, DataAssetArgs
from phidata.utils.log import logger


class LocalAssetArgs(DataAssetArgs):
    pass


class LocalAsset(DataAsset):
    def __init__(self) -> None:
        super().__init__()

    ######################################################
    ## Get FileSystem
    ######################################################

    def _get_fs(self) -> Optional[Any]:
        from pyarrow import fs

        logger.debug("initializing LocalFileSystem")
        return fs.LocalFileSystem()



# from typing import Optional, Literal
#
# from phidata.asset.file import File
# from phidata.asset.aws.s3.object import S3Object
# from phidata.utils.log import logger
# from phidata.task.python_task import PythonTask, PythonTaskArgs
#
#
# class DownloadUrlToS3Args(PythonTaskArgs):
#     url: str
#     s3_object: S3Object
#     # How to behave if the object already exists.
#     # fail: Raise a ValueError.
#     # replace: Replace the object
#     # skip: Do nothing
#     if_exists: Literal["fail", "replace", "skip"] = "replace"
#     # Optional File object to use
#     file: Optional[File] = None
#
#
# class DownloadUrlToS3(PythonTask):
#     def __init__(
#         self,
#         url: str,
#         s3_object: S3Object,
#         if_exists: Literal["fail", "replace", "append"] = "replace",
#         file: Optional[File] = None,
#         name: str = "download_url_to_s3",
#         task_id: Optional[str] = None,
#         dag_id: Optional[str] = None,
#         version: Optional[str] = None,
#         enabled: bool = True,
#     ):
#         super().__init__()
#         try:
#             self.args: DownloadUrlToS3Args = DownloadUrlToS3Args(
#                 url=url,
#                 s3_object=s3_object,
#                 if_exists=if_exists,
#                 file=file,
#                 name=name,
#                 task_id=task_id,
#                 dag_id=dag_id,
#                 version=version,
#                 enabled=enabled,
#                 entrypoint=download_url_to_s3,
#             )
#         except Exception as e:
#             logger.error(f"Args for {self.__class__.__name__} are not valid")
#             raise
#
#     @property
#     def url(self) -> str:
#         return self.args.url
#
#     @url.setter
#     def url(self, url: str) -> None:
#         if url is not None:
#             self.args.url = url
#
#     @property
#     def s3_object(self) -> Optional[S3Object]:
#         return self.args.s3_object
#
#     @s3_object.setter
#     def s3_object(self, s3_object: S3Object) -> None:
#         if s3_object is not None:
#             self.args.s3_object = s3_object
#
#     @property
#     def if_exists(self) -> Literal["fail", "replace", "skip"]:
#         return self.args.if_exists if self.args else "skip"
#
#     @if_exists.setter
#     def if_exists(self, if_exists: Literal["fail", "replace", "skip"]) -> None:
#         if if_exists is not None:
#             self.args.if_exists = if_exists
#
#     @property
#     def file(self) -> Optional[File]:
#         return self.args.file
#
#     @file.setter
#     def file(self, file: File) -> None:
#         if file is not None:
#             self.args.file = file
#
#
# def best_guess_file_name(url: str) -> str:
#
#     from phidata.utils.dttm import get_today_utc_date_str
#
#     fn: Optional[str] = None
#     if "/" in url:
#         fn = url.split("/")[-1]
#     if fn is None:
#         fn = f"tmp__{get_today_utc_date_str()}"
#     logger.debug("fn: {}".format(fn))
#     return fn
#
#
# def download_url_to_s3(**kwargs) -> bool:
#
#     args: DownloadUrlToS3Args = DownloadUrlToS3Args.from_kwargs(kwargs)
#     # logger.info(f"DownloadUrlToS3Args: {args}")
#
#     if args.url is None:
#         logger.error("URL not available")
#         return False
#
#     if args.s3_object is None:
#         logger.error("S3Object not available")
#         return False
#
#     if args.file is None:
#         args.file = File.temporary(best_guess_file_name(args.url))
#
#     download_success = args.file.download_url(args.url)
#     if download_success:
#         return args.s3_object.upload_file(file=args.file, if_exists=args.if_exists)
#     return False



# from pathlib import Path
# from typing import Optional, Literal, Union, List
#
# from phidata.asset.file import File, FileType
# from phidata.table.sql import SqlTable
# from phidata.utils.cli_console import print_info, print_warning
# from phidata.utils.log import logger
# from phidata.task.python_task import PythonTask, PythonTaskArgs
#
#
# class DownloadUrlToSqlArgs(PythonTaskArgs):
#     url: str
#     sql_table: SqlTable
#     # How to behave if the sql_table already exists.
#     # fail: Raise a ValueError.
#     # replace: Drop the table before inserting new values.
#     # append: Insert new values to the existing table.
#     if_exists: Optional[Literal["fail", "replace", "append"]] = None
#     # Write DataFrame index as a column.
#     # Uses index_label as the column name in the table.
#     index: Optional[bool] = None
#     # Column label for index column(s).
#     # If None is given (default) and index is True, then the index names are used.
#     # A sequence should be given if the DataFrame uses MultiIndex.
#     index_label: Optional[Union[str, List[str]]] = None
#     # Specify the number of rows in each batch to be written at a time.
#     # By default, all rows will be written at once.
#     chunksize: Optional[int] = None
#     create_db_engine_from_conn_id: bool = True
#     # Optional File object to use
#     file: Optional[File] = None
#
#
# class DownloadUrlToSql(PythonTask):
#     def __init__(
#         self,
#         url: str,
#         sql_table: SqlTable,
#         if_exists: Optional[Literal["fail", "replace", "append"]] = None,
#         index: Optional[bool] = None,
#         index_label: Optional[Union[str, List[str]]] = None,
#         chunksize: Optional[int] = None,
#         file: Optional[File] = None,
#         name: str = "download_url_to_sql",
#         task_id: Optional[str] = None,
#         dag_id: Optional[str] = None,
#         version: Optional[str] = None,
#         enabled: bool = True,
#     ):
#         super().__init__()
#         try:
#             self.args: DownloadUrlToSqlArgs = DownloadUrlToSqlArgs(
#                 url=url,
#                 sql_table=sql_table,
#                 if_exists=if_exists,
#                 index=index,
#                 index_label=index_label,
#                 chunksize=chunksize,
#                 file=file,
#                 name=name,
#                 task_id=task_id,
#                 dag_id=dag_id,
#                 version=version,
#                 enabled=enabled,
#                 entrypoint=download_url_to_sql,
#             )
#         except Exception as e:
#             logger.error(f"Args for {self.__class__.__name__} are not valid")
#             raise
#
#     @property
#     def url(self) -> str:
#         return self.args.url
#
#     @url.setter
#     def url(self, url: str) -> None:
#         if url is not None:
#             self.args.url = url
#
#     @property
#     def sql_table(self) -> Optional[SqlTable]:
#         return self.args.sql_table
#
#     @sql_table.setter
#     def sql_table(self, sql_table: SqlTable) -> None:
#         if sql_table is not None:
#             self.args.sql_table = sql_table
#
#     @property
#     def file(self) -> Optional[File]:
#         return self.args.file
#
#     @file.setter
#     def file(self, file: File) -> None:
#         if file is not None:
#             self.args.file = file
#
#
# def best_guess_file_name(url: str) -> str:
#
#     from phidata.utils.dttm import get_today_utc_date_str
#
#     fn: Optional[str] = None
#     if "/" in url:
#         fn = url.split("/")[-1]
#     if fn is None:
#         fn = f"tmp__{get_today_utc_date_str()}"
#     logger.debug("fn: {}".format(fn))
#     return fn
#
#
# def download_url_to_file(args: DownloadUrlToSqlArgs) -> bool:
#
#     import httpx
#     import rich.progress
#
#     url = args.url
#     if url is None:
#         logger.error("URL not available")
#         return False
#
#     if args.file is None:
#         args.file = File.temporary(best_guess_file_name(args.url))
#
#     file_path: Optional[Path] = args.file.file_path
#     if file_path is None:
#         logger.error("FilePath not available")
#         return False
#     file_dir: Path = file_path.parent.resolve()
#
#     # Download URL contents
#     print_info("Downloading")
#     print_info(f"  Url: {url}")
#     print_info(f"  To : {file_path}")
#
#     # Create the directory if it does not exist
#     if not file_dir.exists():
#         file_dir.mkdir(parents=True, exist_ok=True)
#
#     bytes_to_download: int = 0
#     try:
#         with file_path.open("w") as open_file:
#             with httpx.stream("GET", url) as response:
#                 try:
#                     headers: httpx.Headers = response.headers
#                     # logger.debug("headers: {}".format(headers))
#                     # logger.debug("status_code: {}".format(response.status_code))
#                     bytes_to_download = int(headers["Content-Length"])
#                     # logger.debug("bytes_to_download : {}".format(bytes_to_download))
#                 except Exception:
#                     print_warning(f"Could not get total bytes_to_download from headers")
#
#                 with rich.progress.Progress(
#                     "[progress.percentage]{task.percentage:>3.0f}%",
#                     rich.progress.BarColumn(50),
#                     rich.progress.DownloadColumn(),
#                     rich.progress.TransferSpeedColumn(),
#                 ) as progress:
#                     download_task = progress.add_task(
#                         "Download", total=bytes_to_download
#                     )
#                     for chunk in response.iter_text():
#                         open_file.write(chunk)
#                         progress.update(
#                             download_task, completed=response.num_bytes_downloaded
#                         )
#         logger.info("Url downloaded")
#         return True
#     except Exception as e:
#         logger.error("Could not download url, please try again")
#         logger.exception(e)
#     return False
#
#
# def load_file_to_sql_table_pandas(args: DownloadUrlToSqlArgs) -> bool:
#
#     import pandas as pd
#
#     file_to_load: Optional[File] = args.file
#     if file_to_load is None:
#         logger.error("File not available")
#         return False
#     file_path: Optional[Path] = file_to_load.file_path
#     if file_path is None:
#         logger.error("FilePath not available")
#         return False
#     file_type: Optional[FileType] = file_to_load.file_type
#     if file_type is None:
#         logger.error("FileType not available")
#         return False
#     # logger.debug("File: {}".format(file_to_load.args))
#
#     sql_table: Optional[SqlTable] = args.sql_table
#     if sql_table is None:
#         logger.error("SqlTable not available")
#         return False
#     # logger.debug("SqlTable: {}".format(sql_table.args))
#
#     logger.info("Reading: {}".format(file_path))
#     df: Optional[pd.DataFrame] = None
#     if file_type == FileType.CSV:
#         df = pd.read_csv(file_path)
#     elif file_type == FileType.JSON:
#         df = pd.read_json(file_path)
#     elif file_type == FileType.TSV:
#         df = pd.read_csv(file_path, sep="\t")
#
#     if df is not None:
#         # logger.info()("DataFrame:\n{}".format(df.head()))
#         logger.info("Writing to table: {}".format(sql_table.name))
#         upload_success = sql_table.write_pandas_df(
#             df=df,
#             if_exists=args.if_exists,
#             index=args.index,
#             index_label=args.index_label,
#             chunksize=args.chunksize,
#         )
#         return upload_success
#     else:
#         logger.info("Could not read file into DataFrame")
#         return False
#
#
# def download_url_to_sql(**kwargs) -> bool:
#
#     args: DownloadUrlToSqlArgs = DownloadUrlToSqlArgs.from_kwargs(kwargs)
#     # logger.info(f"DownloadUrlToSqlArgs: {args}")
#
#     download_success = download_url_to_file(args)
#     if download_success:
#         return load_file_to_sql_table_pandas(args)
#     return False
